#include "mainwindow.h"
#include "ui_mainwindow.h"

#include "GnuPlot.h"



int  GNUPlot::setMatrix(boost::multi_array<std::complex<double>, 3>* matrix, std::vector<int> indexes)
{

    try
    {
        int ret = 0;

        // メインデータの保持
        this->matrix = matrix;

        if (!matrix->num_dimensions() == indexes.size() && !matrix->num_dimensions() == indexes.size() + 1)
        {
            // "マトリックスの次元とインデックスの次元が異なります。(インデックスの次元数は、マトリックスの次元数+1も有効です)"
            qDebug() << "Incompatible matrix's dimension and index's dimension";
            ErrorMsg("Incompatible matrix's dimension and index's dimension");
            return -1;
        }

        // axises のインデックスの初期化
        axises.resize(matrix->num_dimensions() + 1); // +1 はZ軸

        ret = this->setIndexes(indexes);

        return ret;

    }
    catch (const std::exception& e)
    {
         ErrorMsg(e.what());
        return -1;
    }
    catch (...)
    {
        ErrorMsg("unexpected error");
        return -1;
    }


}



//----------------------------------------------//
//
//----------------------------------------------//
void GNUPlot::setStyle(int plot_style, int surface, int color_map)
{
    this->plot_style = plot_style;
    this->color_map  = color_map;
    this->surface    = surface;
}


//----------------------------------------------//
//
//----------------------------------------------//
void GNUPlot::setTitle(std::string title)
{
    this->title = title;
}


//----------------------------------------------//
// 引数の次元は、マトリックスの次元に対応しなければならない
//----------------------------------------------//
void GNUPlot::setLabels(std::vector<std::string> labels)
{
    for (uint i=0; i< labels.size() && i < matrix->num_dimensions() + 1; i++)
    {
        axises[i].label = labels[i];
    }

}



//----------------------------------------------//
// マトリックスの次元とは関係ない
//----------------------------------------------//
void GNUPlot::setCurrentLabels(std::string x, std::string y, std::string z)
{
    axises[x_index].label = x;
    axises[y_index].label = y;
    axises[z_index].label = z;
}



//----------------------------------------------//
// 引数の次元は、マトリックスの次元に対応しなければならない
//----------------------------------------------//
void GNUPlot::setWeights(std::vector<double> weights)
{
    for (uint i=0; i< weights.size() && i < matrix->num_dimensions() + 1; i++)
    {
        axises[i].weight = weights[i];
    }
}



//----------------------------------------------//
// マトリックスの次元とは関係ない
//----------------------------------------------//
void GNUPlot::setCurrentWeight(double x, double y, double z)
{
    axises[x_index].weight = x;
    axises[y_index].weight = y;
    axises[z_index].weight = z;
}


//----------------------------------------------//
// 引数の次元は、マトリックスの次元に対応しなければならない
//----------------------------------------------//
void GNUPlot::setRanges(std::vector<std::pair<std::string, std::string> > ranges)
{
    for (uint i=0; i<ranges.size() && i < matrix->num_dimensions() + 1; i++) {
        axises[i].range = ranges[i];
    }
}



//----------------------------------------------//
// マトリックスの次元とは関係ない
//----------------------------------------------//
void GNUPlot::setCurrentRange(std::pair<std::string, std::string> x,
                     std::pair<std::string, std::string> y,
                     std::pair<std::string, std::string> z)
{
    axises[x_index].range = x;
    axises[y_index].range = y;
    axises[z_index].range = z;
}




//----------------------------------------------//
//
//----------------------------------------------//
int GNUPlot::setIndexes(std::vector<int> indexes)
{

    bool find_x=false, find_y=false, find_z=false;

    for (uint i=0; i < indexes.size(); i++)
    {
        //　軸のインデックスのコピー
        axises[i].index = indexes[i];

        // x軸、y軸の存在チェック
        if (indexes[i] ==  AXIS_X)
        {
            x_index = i;
            find_x = true;
        }
        else if (indexes[i] == AXIS_Y)
        {
            y_index = i;
            find_y = true;
        }
        else if (indexes[i] == AXIS_Z) // Z軸が定義されている場合
        {
            z_index = i;
            find_z = true;
        }
    }

    if (find_x == false || find_y == false || (find_z && (uint)z_index != matrix->num_dimensions()))
    {
        if (find_x == false || find_y == false )
        {
            // "X,Y軸のインデックスが指定されていません"
            ErrorMsg("Axis's index is Undefined on X or Y");
            return -1;
        }
        else
        {
            // "Z軸のインデックスが指定されていますが、無効なインデックスです。"
            ErrorMsg("Z axis's Index is defined, but it is inappropriate");
            return -1;
        }
    }

    // Z軸が未定義の場合
    if (matrix->num_dimensions() == indexes.size())
    {
        z_index = matrix->num_dimensions();
        axises[z_index].index = AXIS_Z;
    }

    return 0;
}







//----------------------------------------------//
//
//----------------------------------------------//
void GNUPlot::setZUnit(int value)
{
    qDebug() << __func__;
    z_unit = value;
}



/*
 * valueの数字を文字列に変換します。
 * 複素数の場合は 実部+虚部i の形式に変換します。
 * complexがfalseの場合は実部のみに変換します。
 */
std::string toString( std::complex<double> value, bool complex=true)
{
    (void)complex;
    std::string str = (boost::format("%1%") % value).str();
    return str;
}





//----------------------------------------------//
//
//----------------------------------------------//
std::string GNUPlot::plot(std::string fullpath)
{

    std::string file = plotToTempFile(fullpath);
    qint64 pid;
    int ret = QProcess::startDetached("gnuplot", QStringList() << "-persist" << QString(file.c_str()), QDir::currentPath(), &pid);

    if (!ret)
    {
          ErrorMsg(QString("Command error: gnuplot --persist %1").arg(file.c_str()));
    }

    return file;
}




//----------------------------------------------//
//
//----------------------------------------------//
std::string GNUPlot::plotToTempFile(std::string pathname)
{

    // テンポラリファイル
    char fname[256];
    sprintf(fname, "%s/temp-XXXXXX", pathname.c_str());
    int ret = mkstemp(fname);
    (void)ret;
    std::ofstream ofs(fname);

    if(ofs.fail())
    {
        ErrorMsg(  QString( "Can't open file : %1" ).arg(fname));
    }
    plotToStream(ofs);

    return fname;
}




//----------------------------------------------//
//
//----------------------------------------------//
void GNUPlot::plotToFile(std::string fname)
{

    std::ofstream ofs(fname.c_str());
    if(ofs.fail())
    {
        ErrorMsg(QString("Can't open file : ")+  QString(fname.c_str())   );
    }
    plotToStream(ofs);

}




//----------------------------------------------//
//
//----------------------------------------------//
void GNUPlot::exportCSVToFile(std::string fname,   bool complex)
//void GNUPlot::exportCSVToFile(std::string fname,   bool complex=true)
{

    std::ofstream ofs(fname.c_str());
    if(ofs.fail())
    {
        ErrorMsg(QString("Can't open file : ")+ QString(fname.c_str()) );
    }
    exportCSVToStream(ofs, complex);
}




//----------------------------------------------//
/*
 * X軸方向をテキストの行,Y軸方向をテキストの列に対応するようにして、CSV形式でそのまま出力します
 * 各軸のWeightは考慮されません
 * complexがfalseの場合は、データが複素数でも実部のみ出力します
 */
//----------------------------------------------//
void GNUPlot::exportCSVToStream(std::ostream& os, bool complex)
{

    boost::array<int, 3> index;
    // プロットしない固定次元をindex配列に指定する
    for (uint i=0; i < axises.size() - 1; i++)
    {
        index[i] = axises[i].index;
    }

    // プロットする２次元分の処理
    for (uint y = 0; y < matrix->shape()[y_index]; y++)
    {
        for (uint x = 0; x < matrix->shape()[x_index]; x++)
        {
            index[x_index] = x;
            index[y_index] = y;

            if (x)
            {
                os <<  ",";
            }
            os  << toString((*matrix)(index), complex);
        }
        os <<  "\n";
    }
}





//----------------------------------------------//
//
//----------------------------------------------//
void GNUPlot::plotToStream(std::ostream& os)
{
    try
    {

        // タイトル
        os << boost::format("set title '%s'\n") %title;

        // 表示範囲
        os << boost::format("set xrange [%s:%s]\n") %axises[x_index].range.first %axises[x_index].range.second;
        os << boost::format("set yrange [%s:%s]\n") %axises[y_index].range.first %axises[y_index].range.second;
        os << boost::format("set zrange [%s:%s]\n") %axises[z_index].range.first %axises[z_index].range.second;

        // 軸のタイトル
        os << boost::format("set xlabel '%s'\n") %axises[x_index].label;
        os << boost::format("set ylabel '%s'\n") %axises[y_index].label;
        os << boost::format("set zlabel '%s'\n") %axises[z_index].label;

        // XYZ軸
        os << "set xzeroaxis lt -1\n";
        os << "set yzeroaxis lt -1\n";
        os << "set zeroaxis lt -1\n";
        os << "set ticslevel 0\n"; // Z軸交点固定

        // カラーマップ
        {
            std::string color_map_string;
            switch(color_map)
            {
                case ColorMap_PM3D:
                    color_map_string = "pm3d";
                    break;
                case ColorMap_PM3D_MAP:
                    color_map_string = "pm3d map";
                    break;
                case ColorMap_PM3D_AT_B:
                    color_map_string = "pm3d at b";
                    break;
                default:
                    color_map_string = "";
                    break;
            }
            if (color_map_string != "")
                os <<  boost::format("set %s\n") %color_map_string;
        }


        // プロットスタイル
        {
            std::string style_string;
            switch(plot_style) {
                case PlotStyle_UNDEFINE:
                    if (color_map == ColorMap_PM3D) {
                        style_string = "with pm3d";
                    } else {
                        style_string = "";
                    }
                    break;
                case PlotStyle_DOTS:
                    style_string = "with dots";
                    break;
                case PlotStyle_POINTS:
                    style_string = "with points";
                    break;
                case PlotStyle_LINES:
                    style_string = "with lines";
                    break;
                case PlotStyle_LINESPOINTS:
                    style_string = "with linespoints";
            }
            os << boost::format("splot '-' notitle %s\n") %style_string;
        }


        boost::array<int, 3> index;


        // プロットしない固定次元をindex配列に指定する
        for (uint i=0; i < axises.size() - 1; i++)
        {
            index[i] = axises[i].index;
        }

        // プロットする２次元分の処理
        for (uint y = 0; y < matrix->shape()[y_index]; y++)// このループはlines の時に x-z平面方向の線を引くようにしている。
        {
            for (uint x = 0; x < matrix->shape()[x_index]; x++)
            {
                index[x_index] = x;
                index[y_index] = y;
                os << boost::format("%f %f %f\n") %(x * axises[x_index].weight) %(y * axises[y_index].weight) %( GNUPlot::calcZValue( z_unit, (*matrix)(index)) *axises[z_index].weight);
            }

            // Surface
            {
                std::string surface_string;
                switch(surface) {
                    case Surface_MESH:
                        surface_string = "";
                        break;
                    case Surface_LINE:
                        surface_string = "\n";
                        break;
                    default:
                        surface_string = "";
                        break;
                }
                os <<  boost::format("%s\n") %surface_string;
            }
        }
        os <<  "e\n";
        os <<  "pause -1\n";

        return;

    }
    catch (const std::exception& e)
    {
         ErrorMsg(e.what());
    }
    catch (...)
    {
        ErrorMsg(("unexpected error"));
    }


}



double GNUPlot::calcZValue(int z_unit, std::complex<double>value)
{

    double re = value.real();
    double im = value.imag();
    switch(z_unit)
    {
        case DISP_REAL: //3
            return re;
            break;
        case DISP_AMP:
        return sqrt(re*re + im*im);
            break;
        case DISP_db:
            return log10(sqrt(re*re + im*im)) * 10;
            break;
        case DISP_PHASE:
            return atan2(im, re) * 180/M_PI;
            break;
        default:
            GNUPlot::ErrorMsg(QString("Unknown Z Axis's Unit:%1").arg(z_unit));
        break;
    }
}



void GNUPlot::ErrorMsg(QString msg)
{

    QMessageBox  msgBox;
    msgBox.setIcon(QMessageBox::Critical);
    msgBox.setText(msg);
    msgBox.exec();
}



